/*
 * This file is part of the GROMACS molecular simulation package.
 *
 * Copyright 2015- The GROMACS Authors
 * and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
 * Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
 *
 * GROMACS is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * GROMACS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with GROMACS; if not, see
 * https://www.gnu.org/licenses, or write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * If you want to redistribute modifications to GROMACS, please
 * consider that scientific software is very special. Version
 * control is crucial - bugs must be traceable. We will be happy to
 * consider code for inclusion in the official distribution, but
 * derived work must not be called official GROMACS. Details are found
 * in the README & COPYING files - if they are missing, get the
 * official version at https://www.gromacs.org.
 *
 * To help us fund GROMACS development, we humbly ask that you cite
 * the research papers on the package. Check out https://www.gromacs.org.
 */

/*! \internal \file
 * \brief
 * Implements the one method of the PointState class called only for one point per step.
 *
 * \author Viveca Lindahl
 * \author Berk Hess <hess@kth.se>
 * \ingroup module_awh
 */

#include "gmxpre.h"

#include "pointstate.h"

namespace gmx
{

namespace
{

/*! \brief Returns the exponent c where exp(c) = exp(a) + exp(b).
 *
 * \param[in] a  First exponent.
 * \param[in] b  Second exponent.
 * \returns c.
 */
double expSum(double a, double b)
{
    return (a > b ? a : b) + std::log1p(std::exp(-std::fabs(a - b)));
}

} // namespace

void PointState::samplePmf(double convolvedBias)
{
    if (inTargetRegion())
    {
        logPmfSum_ = expSum(logPmfSum_, -convolvedBias);
        numVisitsIteration_ += 1;
    }
}

void PointState::updatePmfUnvisited(double bias)
{
    if (inTargetRegion())
    {
        logPmfSum_ = expSum(logPmfSum_, -bias);
    }
}

} // namespace gmx
