/*
 * This file is part of the GROMACS molecular simulation package.
 *
 * Copyright 2018- The GROMACS Authors
 * and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
 * Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
 *
 * GROMACS is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * GROMACS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with GROMACS; if not, see
 * https://www.gnu.org/licenses, or write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * If you want to redistribute modifications to GROMACS, please
 * consider that scientific software is very special. Version
 * control is crucial - bugs must be traceable. We will be happy to
 * consider code for inclusion in the official distribution, but
 * derived work must not be called official GROMACS. Details are found
 * in the README & COPYING files - if they are missing, get the
 * official version at https://www.gromacs.org.
 *
 * To help us fund GROMACS development, we humbly ask that you cite
 * the research papers on the package. Check out https://www.gromacs.org.
 */

#include "gmxpre.h"

#include "gmxapi/version.h"

#include <string>

namespace gmxapi
{

version_t Version::majorVersion()
{
    return c_majorVersion;
}

version_t Version::minorVersion()
{
    return c_minorVersion;
}

version_t Version::patchVersion()
{
    return c_patchVersion;
}

std::string Version::release()
{
    return c_release;
}

bool Version::hasFeature(const std::string& featurename)
{
    // For features introduced without an incompatible API change or where
    // semantic versioning is otherwise insufficient, we can consult a map, TBD.
    (void)featurename;
    return false;
}

bool Version::isAtLeast(version_t major, version_t minor, version_t patch)
{
    if (Version::majorVersion() < major)
    {
        return false;
    }
    if (Version::majorVersion() > major)
    {
        return true;
    }
    if (Version::minorVersion() < minor)
    {
        return false;
    }
    if (Version::minorVersion() > minor)
    {
        return true;
    }
    return Version::patchVersion() >= patch;
}


} // end namespace gmxapi
