#define TESTNAME "Initial coding. Intra frame triple algorithm. Cubic cell."
#define FILENAME "test1.tng_compress"
#define ALGOTEST
#define NATOMS 1000
#define CHUNKY 1
#define SCALE 0.1
#define PRECISION 0.01
#define WRITEVEL 0
#define VELPRECISION 0.1
#define INITIALCODING 3
#define INITIALCODINGPARAMETER -1
#define CODING 1
#define CODINGPARAMETER 0
#define VELCODING 0
#define VELCODINGPARAMETER 0
#define INTMIN1 0
#define INTMIN2 0
#define INTMIN3 0
#define INTMAX1 10000
#define INTMAX2 10000
#define INTMAX3 10000
#define NFRAMES 1000
#define EXPECTED_FILESIZE 2776230.
