/*
 * This file is part of the GROMACS molecular simulation package.
 *
 * Copyright 2018- The GROMACS Authors
 * and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
 * Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
 *
 * GROMACS is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * GROMACS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with GROMACS; if not, see
 * https://www.gnu.org/licenses, or write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * If you want to redistribute modifications to GROMACS, please
 * consider that scientific software is very special. Version
 * control is crucial - bugs must be traceable. We will be happy to
 * consider code for inclusion in the official distribution, but
 * derived work must not be called official GROMACS. Details are found
 * in the README & COPYING files - if they are missing, get the
 * official version at https://www.gromacs.org.
 *
 * To help us fund GROMACS development, we humbly ask that you cite
 * the research papers on the package. Check out https://www.gromacs.org.
 */
/*! \internal
 * \brief Defines the dispatch function for the .mdp integrator field.
 *
 * \author David van der Spoel <david.vanderspoel@icm.uu.se>
 * \author Mark Abraham <mark.j.abraham@gmail.com>
 * \ingroup module_mdrun
 */
#include "gmxpre.h"

#include "legacysimulator.h"

#include "gromacs/mdtypes/inputrec.h"
#include "gromacs/mdtypes/md_enums.h"
#include "gromacs/utility/exceptions.h"

namespace gmx
{
//! \brief Run the correct integrator function.
void LegacySimulator::run()
{
    switch (inputRec_->eI)
    {
        case IntegrationAlgorithm::MD:
        case IntegrationAlgorithm::BD:
        case IntegrationAlgorithm::SD1:
        case IntegrationAlgorithm::VV:
        case IntegrationAlgorithm::VVAK:
            if (!EI_DYNAMICS(inputRec_->eI))
            {
                GMX_THROW(APIError(
                        "do_md integrator would be called for a non-dynamical integrator"));
            }
            if (doRerun_)
            {
                do_rerun();
            }
            else
            {
                do_md();
            }
            break;
        case IntegrationAlgorithm::Mimic:
            if (doRerun_)
            {
                do_rerun();
            }
            else
            {
                do_mimic();
            }
            break;
        case IntegrationAlgorithm::Steep: do_steep(); break;
        case IntegrationAlgorithm::CG: do_cg(); break;
        case IntegrationAlgorithm::NM: do_nm(); break;
        case IntegrationAlgorithm::LBFGS: do_lbfgs(); break;
        case IntegrationAlgorithm::TPI:
        case IntegrationAlgorithm::TPIC:
            if (!EI_TPI(inputRec_->eI))
            {
                GMX_THROW(APIError("do_tpi integrator would be called for a non-TPI integrator"));
            }
            do_tpi();
            break;
        case IntegrationAlgorithm::SD2Removed:
            GMX_THROW(NotImplementedError("SD2 integrator has been removed"));
        default: GMX_THROW(APIError("Non existing integrator selected"));
    }
}

} // namespace gmx
