#define TESTNAME "Position coding. XTC3 algorithm. High accuracy. Cubic cell."
#define FILENAME "test49.tng_compress"
#define ALGOTEST
#define NATOMS 100000
#define CHUNKY 10
#define SCALE 0.5
#define PRECISION 1e-8
#define WRITEVEL 0
#define VELPRECISION 0.1
#define INITIALCODING 3
#define INITIALCODINGPARAMETER -1
#define CODING 10
#define CODINGPARAMETER 0
#define VELCODING 4
#define VELCODINGPARAMETER 0
#define INTMIN1 0
#define INTMIN2 0
#define INTMIN3 0
#define INTMAX1 805306368
#define INTMAX2 805306368
#define INTMAX3 805306368
#define NFRAMES 20
#define EXPECTED_FILESIZE 3585605.
