/*
 * This file is part of the GROMACS molecular simulation package.
 *
 * Copyright 2012- The GROMACS Authors
 * and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
 * Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
 *
 * GROMACS is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * GROMACS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with GROMACS; if not, see
 * https://www.gnu.org/licenses, or write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * If you want to redistribute modifications to GROMACS, please
 * consider that scientific software is very special. Version
 * control is crucial - bugs must be traceable. We will be happy to
 * consider code for inclusion in the official distribution, but
 * derived work must not be called official GROMACS. Details are found
 * in the README & COPYING files - if they are missing, get the
 * official version at https://www.gromacs.org.
 *
 * To help us fund GROMACS development, we humbly ask that you cite
 * the research papers on the package. Check out https://www.gromacs.org.
 */
/*! \libinternal \file
 * \brief
 * Declares mock implementation of gmx::IHelpTopic.
 *
 * \author Teemu Murtola <teemu.murtola@gmail.com>
 * \inlibraryapi
 * \ingroup module_onlinehelp
 */
#ifndef GMX_ONLINEHELP_TESTS_MOCK_HELPTOPIC_H
#define GMX_ONLINEHELP_TESTS_MOCK_HELPTOPIC_H

#include <string>

#include <gmock/gmock.h>

#include "gromacs/onlinehelp/helptopic.h"
#include "gromacs/onlinehelp/helpwritercontext.h"

namespace gmx
{
namespace test
{

class MockHelpTopic : public AbstractCompositeHelpTopic
{
public:
    static MockHelpTopic& addSubTopic(gmx::AbstractCompositeHelpTopic* parent,
                                      const char*                      name,
                                      const char*                      title,
                                      const char*                      text);

    MockHelpTopic(const char* name, const char* title, const char* text);
    ~MockHelpTopic() override;

    const char* name() const override;
    const char* title() const override;

    MOCK_CONST_METHOD1(writeHelp, void(const HelpWriterContext& context));

    MockHelpTopic& addSubTopic(const char* name, const char* title, const char* text);
    using AbstractCompositeHelpTopic::addSubTopic;

    /*! \brief
     * Calls base class writeHelp() method.
     *
     * This provides the possibility for the mock to do the actual help
     * writing.
     */
    void writeHelpBase(const HelpWriterContext& context)
    {
        AbstractCompositeHelpTopic::writeHelp(context);
    }

private:
    std::string helpText() const override;

    const char* name_;
    const char* title_;
    const char* text_;
};

} // namespace test
} // namespace gmx

#endif
