/*
 * This file is part of the GROMACS molecular simulation package.
 *
 * Copyright 2010- The GROMACS Authors
 * and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
 * Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
 *
 * GROMACS is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * GROMACS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with GROMACS; if not, see
 * https://www.gnu.org/licenses, or write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * If you want to redistribute modifications to GROMACS, please
 * consider that scientific software is very special. Version
 * control is crucial - bugs must be traceable. We will be happy to
 * consider code for inclusion in the official distribution, but
 * derived work must not be called official GROMACS. Details are found
 * in the README & COPYING files - if they are missing, get the
 * official version at https://www.gromacs.org.
 *
 * To help us fund GROMACS development, we humbly ask that you cite
 * the research papers on the package. Check out https://www.gromacs.org.
 */
/*! \internal \file
 * \brief
 * Implements registerTrajectoryAnalysisModules().
 *
 * \author Teemu Murtola <teemu.murtola@gmail.com>
 * \ingroup module_trajectoryanalysis
 */
#include "gmxpre.h"

#include "modules.h"

#include "modules/angle.h"
#include "modules/convert_trj.h"
#include "modules/distance.h"
#include "modules/dssp.h"
#include "modules/extract_cluster.h"
#include "modules/freevolume.h"
#include "modules/gyrate.h"
#include "modules/hbond.h"
#include "modules/msd.h"
#include "modules/pairdist.h"
#include "modules/rdf.h"
#include "modules/sasa.h"
#include "modules/scattering.h"
#include "modules/select.h"
#include "modules/trajectory.h"

#include "gromacs/commandline/cmdlinemodulemanager.h"
#include "gromacs/trajectoryanalysis/cmdlinerunner.h"

namespace gmx
{

namespace
{

/*! \brief
 * Convenience method for registering a command-line module for trajectory
 * analysis.
 *
 * \tparam ModuleInfo  Info about trajectory analysis module to wrap.
 *
 * \p ModuleInfo should have static public members
 * `const char name[]`, `const char shortDescription[]`, and
 * `gmx::TrajectoryAnalysisModulePointer create()`.
 *
 * \ingroup module_trajectoryanalysis
 */
template<class ModuleInfo>
void registerModule(CommandLineModuleManager* manager, CommandLineModuleGroup group)
{
    TrajectoryAnalysisCommandLineRunner::registerModule(
            manager, ModuleInfo::name, ModuleInfo::shortDescription, &ModuleInfo::create);
    group.addModule(ModuleInfo::name);
}

} // namespace

//! \cond libapi
void registerTrajectoryAnalysisModules(CommandLineModuleManager* manager)
{
    using namespace gmx::analysismodules;
    CommandLineModuleGroup group = manager->addModuleGroup("Trajectory analysis");
    registerModule<AngleInfo>(manager, group);
    registerModule<ConvertTrjInfo>(manager, group);
    registerModule<DistanceInfo>(manager, group);
    registerModule<DsspInfo>(manager, group);
    registerModule<ExtractClusterInfo>(manager, group);
    registerModule<FreeVolumeInfo>(manager, group);
    registerModule<HbondInfo>(manager, group);
    registerModule<MsdInfo>(manager, group);
    registerModule<PairDistanceInfo>(manager, group);
    registerModule<RdfInfo>(manager, group);
    registerModule<SasaInfo>(manager, group);
    registerModule<ScatteringInfo>(manager, group);
    registerModule<SelectInfo>(manager, group);
    registerModule<TrajectoryInfo>(manager, group);
    registerModule<GyrateInfo>(manager, group);
}
//! \endcond

} // namespace gmx
