#ifndef _LAPACK_LIMITS_H_
#define _LAPACK_LIMITS_H_

#define DSTEBZ_BLOCKSIZE 1

#define DORGBR_BLOCKSIZE 32
#define DORGBR_MINBLOCKSIZE 2
#define DORGBR_CROSSOVER 128

#define DGEQRF_BLOCKSIZE 32
#define DGEQRF_MINBLOCKSIZE 2
#define DGEQRF_CROSSOVER 128

#define DORGQR_BLOCKSIZE 32
#define DORGQR_MINBLOCKSIZE 2
#define DORGQR_CROSSOVER 128

#define DORMLQ_BLOCKSIZE 32
#define DORMLQ_MINBLOCKSIZE 2
#define DORMLQ_CROSSOVER 128

#define DORMQL_BLOCKSIZE 32
#define DORMQL_MINBLOCKSIZE 2
#define DORMQL_CROSSOVER 128

#define DSYTRD_BLOCKSIZE 32
#define DSYTRD_MINBLOCKSIZE 2
#define DSYTRD_CROSSOVER 128

#define DGEBRD_BLOCKSIZE 32
#define DGEBRD_MINBLOCKSIZE 2
#define DGEBRD_CROSSOVER 128

#define DORMQR_BLOCKSIZE 32
#define DORMQR_MINBLOCKSIZE 2
#define DORMQR_CROSSOVER 128

#define DGELQF_BLOCKSIZE 32
#define DGELQF_MINBLOCKSIZE 2
#define DGELQF_CROSSOVER 128

#define DGETRF_BLOCKSIZE 64
#define DGETRF_MINBLOCKSIZE 2

#define DGETRI_BLOCKSIZE 64
#define DGETRI_MINBLOCKSIZE 2

#define DTRTRI_BLOCKSIZE 64

#define DBDSDC_SMALLSIZE 25

#define DBDSQR_MAXITR 6


#endif /* _LAPACK_LIMITS_H_ */
