/*!
\internal \file
\brief
Doxygen documentation file for directories in the source tree.

\author Teemu Murtola <teemu.murtola@gmail.com>
*/

/*! \libinternal
\dir docs
\brief Build system and source code for various documentation items.
 */
/*! \libinternal
\dir docs/doxygen
\brief Doxygen build system and general documentation content.
 */

/*!
\dir src
\brief Main source code directory.
 */

/*!
\dir src/gromacs
\brief Source code for building the `libgromacs` library.
 */

/*!
\dir src/gromacs/analysisdata
\brief \ref module_analysisdata

\ingroup module_analysisdata
 */
/*! \libinternal
\dir src/gromacs/analysisdata/tests
\brief Unit tests for \ref module_analysisdata

\ingroup module_analysisdata
 */

/*!
\dir src/gromacs/commandline
\brief \ref module_commandline

\ingroup module_commandline
 */
/*! \libinternal
\dir src/gromacs/commandline/tests
\brief Unit tests for \ref module_commandline

\ingroup module_commandline
 */

/*! \libinternal
\dir src/gromacs/compat
\brief \ref group_compatibility

\ingroup group_compatibility
 */

/*! \internal
\dir src/gromacs/compat/tests
\brief Tests for \ref group_compatibility

\ingroup group_compatibility
 */

/*! \libinternal
\dir src/gromacs/onlinehelp
\brief \ref module_onlinehelp

\ingroup module_onlinehelp
 */
/*! \libinternal
\dir src/gromacs/onlinehelp/tests
\brief Unit tests for \ref module_onlinehelp

\ingroup module_onlinehelp
 */

/*!
\dir src/gromacs/options
\brief \ref module_options

\ingroup module_options
 */
/*! \libinternal
\dir src/gromacs/options/tests
\brief Unit tests for \ref module_options

\ingroup module_options
 */

/*!
\dir src/gromacs/selection
\brief \ref module_selection

\ingroup module_selection
 */
/*! \libinternal
\dir src/gromacs/selection/tests
\brief Unit tests for \ref module_selection

\ingroup module_selection
 */

/*! \libinternal
\dir src/gromacs/simd
\brief \ref module_simd

\ingroup module_simd
 */
/*! \libinternal
\dir src/gromacs/simd/tests
\brief Unit tests for \ref module_simd

\ingroup module_simd
 */

/*!
\dir src/gromacs/trajectoryanalysis
\brief \ref module_trajectoryanalysis

\ingroup module_trajectoryanalysis
 */
/*! \libinternal
\dir src/gromacs/trajectoryanalysis/tests
\brief Unit tests for \ref module_trajectoryanalysis

\ingroup module_trajectoryanalysis
 */

/*!
\dir src/gromacs/utility
\brief \ref module_utility

\ingroup module_utility
 */
/*! \libinternal
\dir src/gromacs/utility/tests
\brief Unit tests for \ref module_utility

\ingroup module_utility
 */

/*!
\libinternal
\dir src/programs
\brief Source code for building executables, see \ref page_wrapperbinary.
 */
/*!
\libinternal
\dir src/programs/mdrun
\brief Source code specific to `mdrun`.
 */

/*!
\libinternal
\dir src/testutils
\brief \ref module_testutils

\ingroup module_testutils
 */
/*!
\libinternal
\dir src/testutils/tests
\brief Unit tests for \ref module_testutils

\ingroup module_testutils
 */

/*!
\dir share
\brief Directory that contains installed data files.
 */
/*!
\dir share/template
\brief Template code for writing analysis programs.
 */
