
void barrier_data_tester(void);

void bcast_data_tester(void);

void gather_data_tester(void);

void gatherv_data_tester(void);

void scatter_data_tester(void);

void scatterv_data_tester(void);

void alltoallv_data_tester(void);
