
#define TEST_AFFINITY
#define TEST_BARRIERS
#define TEST_COND_VARS
#define TEST_ATOMICS
#define TEST_P2P
#define TEST_MULTICAST
#define TEST_REDUCE
#define TEST_SPLIT

/*#define SHORT_TEST */

#ifndef SHORT_TEST

#define NSPLITTEST 100
#define SEND_RECV_INITIAL_NTRIES 10
#define SEND_RECV_INITIAL_TRYSIZE 5000
#define NTRIES 75
#define TRYSIZE 10000

#else

#define NSPLITTEST 10
#define SEND_RECV_INITIAL_NTRIES 10
#define SEND_RECV_INITIAL_TRYSIZE 50
#define NTRIES 75
#define TRYSIZE 100

#endif




#define SEND_RECV_NTRIES    NTRIES
#define SEND_RECV_TRYSIZE   TRYSIZE
#define BARRIER_NTRIES      NTRIES
#define BARRIER_TRYSIZE     TRYSIZE
#define BCAST_NTRIES        NTRIES
#define BCAST_TRYSIZE       TRYSIZE
#define GATHER_NTRIES       NTRIES
#define GATHER_TRYSIZE      TRYSIZE
#define GATHERV_NTRIES      NTRIES
#define GATHERV_TRYSIZE     TRYSIZE
#define SCATTER_NTRIES      NTRIES
#define SCATTER_TRYSIZE     TRYSIZE
#define SCATTERV_NTRIES     NTRIES
#define SCATTERV_TRYSIZE    TRYSIZE
#define ALLTOALLV_NTRIES    NTRIES
#define ALLTOALLV_TRYSIZE   TRYSIZE
#define REDUCE_NTRIES       NTRIES
#define REDUCE_TRYSIZE      TRYSIZE
