#define TESTNAME "Velocity coding. Interframe triple algorithm. High accuracy. Cubic cell."
#define FILENAME "test54.tng_compress"
#define ALGOTEST
#define NATOMS 100000
#define CHUNKY 10
#define SCALE 0.5
#define PRECISION 1e-8
#define WRITEVEL 1
#define VELINTMUL 100000
#define VELPRECISION 1e-8
#define INITIALCODING 3
#define INITIALCODINGPARAMETER -1
#define CODING 5
#define CODINGPARAMETER 0
#define VELCODING 2
#define VELCODINGPARAMETER -1
#define INTMIN1 0
#define INTMIN2 0
#define INTMIN3 0
#define INTMAX1 805306368
#define INTMAX2 805306368
#define INTMAX3 805306368
#define NFRAMES 100
#define EXPECTED_FILESIZE 161798573.
