#define TESTNAME "Coding. Compress zeros test. positions xtc3"
#define FILENAME "test69.tng_compress"
#define ALGOTEST
#define NATOMS 100
#define CHUNKY 100
#define SCALE 0.
#define PRECISION 0.01
#define WRITEVEL 0
#define VELPRECISION 0.1
#define INITIALCODING 10
#define INITIALCODINGPARAMETER 0
#define CODING 10
#define CODINGPARAMETER 0
#define INITIALVELCODING 3
#define INITIALVELCODINGPARAMETER -1
#define VELCODING 3
#define VELCODINGPARAMETER -1
#define INTMIN1 0
#define INTMIN2 0
#define INTMIN3 0
#define INTMAX1 0
#define INTMAX2 0
#define INTMAX3 0
#define NFRAMES 100
#define EXPECTED_FILESIZE 867.
